class Callback:
    def __init__(self):
        self._observers:list = []

    def callback(self):
        for callback in self._observers:
            #print('announcing change')
            callback()

    def bind_to(self, callback):
        #print(f'bound {id(self)}')
        self._observers.append(callback)



from datetime import datetime
from .thresholdbase import ThresholdBase


class ThresholdLite(ThresholdBase):
    def __init__(self, hub):
        self._hub = hub
        super().__init__(hub)

    @property
    def allowedcurrent(self) -> int:
        amps = self._setcurrentdict()
        ret = 6
        if self._hub.sensors.charger_enabled.value is False or self._hub.sensors.charger_done.value is True:
            return ret
        currents = amps
        for key, value in currents.items():
            if (((key / 60) * (60 - datetime.now().minute) + self._hub.sensors.totalhourlyenergy.value * 1000) / 1000) < self._hub.current_peak_dynamic:
                ret = value
                break
        return ret
    
    async def async_allowed_current(self) -> int:
        amps = await self.async_setcurrentdict()
        ret = 6
        if self._hub.sensors.charger_enabled.value is False or self._hub.sensors.charger_done.value is True:
            return ret
        currents = amps
        for key, value in currents.items():
            if (((key / 60) * (60 - datetime.now().minute) + self._hub.sensors.totalhourlyenergy.value * 1000) / 1000) < self._hub.current_peak_dynamic:
                ret = value
                break
        return ret



from .models.stop_string import AllowanceObj, set_allowance_obj
from .models.datetime_model import DateTimeModel
from .models.hour_price import HourPrice
from .models.list_type import ListType
from .models.hourselection_model import HourSelectionModel
from statistics import stdev, mean
from datetime import date, datetime
from ...models.hourselection.hourselection_options import HourSelectionOptions
from .hourselection_calculations import normalize_prices
from .offset_dict import get_offset_dict, set_offset_dict
from .permittance import set_initial_permittance, set_scooped_permittance
from .max_min_charge import MaxMinCharge


class HourSelectionService:
    def __init__(self, options: HourSelectionOptions = HourSelectionOptions()):
        self.options = options
        self.dtmodel = DateTimeModel()
        self.model = HourSelectionModel()
        self.max_min = MaxMinCharge(service=self, min_price=self.options.min_price)
        self._offset_dict: dict[datetime, dict] = {}

    def update(self):
        for hp in self.model.hours_prices:
            hp.set_passed(self.dtmodel)

    async def async_update(self):
        self.update()

    @property
    def all_hours(self) -> list[HourPrice]:
        self.update()
        return self.model.hours_prices

    @property
    def future_hours(self) -> list[HourPrice]:
        self.update()
        for r in self.model.hours_prices:
            print(r.dt, r.passed, r.permittance)
        ret = [hp for hp in self.model.hours_prices if not hp.passed]
        # if self.max_min.active:
        #     for r in ret:
        #         r.permittance = 0.0 if r.dt in self.max_min.non_hours else 1.0
        return ret

    @property
    def passed_hours(self) -> list[HourPrice]:
        self.update()
        return [hp for hp in self.model.hours_prices if hp.passed]

    @property
    def stopped_string(self) -> str:
        self.update()
        return self.allowance.display_name

    @property
    def allowance(self) -> AllowanceObj:
        return set_allowance_obj(self.dtmodel, self.future_hours)

    @property
    def average_kwh_price(self) -> float:
        return round(
            mean(
                [
                    hp.permittance * hp.price
                    for hp in self.future_hours
                    if hp.permittance > 0
                ]
            ),
            2,
        )

    @property
    def offset_dict(self) -> dict:
        return get_offset_dict(self._offset_dict)

    async def async_update_prices(
        self, prices: list[float], prices_tomorrow: list[float] = []
    ):
        self.model.prices_today = prices  # clean first
        self.model.prices_tomorrow = prices_tomorrow  # clean first
        if self._do_recalculate_prices(prices, prices_tomorrow):
            self.model.hours_prices = await self.async_create_prices(
                prices, prices_tomorrow
            )

    async def async_update_adjusted_average(self, adjusted_average: float) -> None:
        self.model.adjusted_average = adjusted_average
        self.update()
        if len(self.model.hours_prices) > 0:
            await self.async_set_permittance(self.model.hours_prices)

    def _do_recalculate_prices(self, prices, prices_tomorrow) -> bool:
        if [
            hp.price for hp in self.model.hours_prices if hp.day == self.dtmodel.hdate
        ] == prices and len(prices_tomorrow) < 1:
            return False
        return True

    async def async_create_prices(
        self, prices: list[float], prices_tomorrow: list[float] = []
    ) -> list:
        # todo: fix to allow 23, 24,25, 92, 96, 100 for dst-dates.
        match len(prices):
            case 23 | 24 | 25:
                return await self.async_create_hour_prices(
                    prices, prices_tomorrow, False
                )
            case 92 | 96 | 100:
                return await self.async_create_hour_prices(
                    prices, prices_tomorrow, True
                )
            case 0:
                return []
        raise ValueError(
            f"Length of pricelist must be either 23,24,25,92,96 or 100. Your length is {len(prices)}"
        )

    def _is_passed(self, datum, hour, quarter) -> bool:
        if datum == self.dtmodel.hdate_tomorrow:
            return False
        if self.dtmodel.hour > hour:
            return True
        elif self.dtmodel.hour == hour:
            if self.dtmodel.quarter > quarter:
                return True
        return False

    async def async_create_hour_prices(
        self,
        prices: list[float],
        prices_tomorrow: list[float] = [],
        is_quarterly: bool = False,
    ) -> list:
        # todo: handle here first if prices or prices_tomorrow are 92 or 100 in length (dst shift)
        ret = []
        ret.extend(self._set_hourprice_list(prices, is_quarterly, self.dtmodel.hdate))
        ret.extend(
            self._set_hourprice_list(
                prices_tomorrow, is_quarterly, self.dtmodel.hdate_tomorrow
            )
        )
        return await self.async_set_permittance(ret)

    def _set_hourprice_list(
        self, prices: list, is_quarterly: bool, datum: date
    ) -> list[HourPrice]:
        ret = []
        for idx, p in enumerate(prices):
            assert isinstance(p, (float, int))
            hour = int(idx / 4) if is_quarterly else idx
            quarter = idx % 4 if is_quarterly else 0
            ret.append(
                HourPrice(
                    day=datum,
                    hour=hour,
                    quarter=quarter,
                    price=p,
                    passed=self._is_passed(datum, hour, quarter),
                    hour_type=HourPrice.set_hour_type(
                        self.options.absolute_top_price, self.options.min_price, p
                    ),
                    list_type=ListType.Quarterly if is_quarterly else ListType.Hourly,
                )
            )
        return ret

    async def async_set_permittance(
        self, hour_prices: list[HourPrice]
    ) -> list[HourPrice]:
        prices = normalize_prices([hp.price for hp in hour_prices])
        hours_permitted = set_scooped_permittance(
            set_initial_permittance(
                hour_prices,
                self._set_price_mean(prices, self.model.adjusted_average),
                stdev(prices),
            ),
            self.options.cautionhour_type_enum,
        )
        self._offset_dict = set_offset_dict(prices, hours_permitted[0].day)
        return hours_permitted

    @staticmethod
    def _set_price_mean(prices: list[float], adjusted_average: float | None) -> float:
        # print(f"adj: {adjusted_average}")
        if not adjusted_average:
            return mean(prices)
        return mean([adjusted_average, mean(prices)])



from __future__ import annotations
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from ..hoursselection import Hoursselection

import logging
from datetime import datetime
from typing import Tuple
import statistics as stat

from .hoursselection_helpers import async_create_dict
from .hourselection_calculations import (
    normalize_prices,
    async_create_cautions,
    async_get_offset_dict,
    async_should_be_cautionhour,
    async_set_charge_allowance,
)
from ....models.hourselection.hourobjects.hourobject import HourObject
from ....models.hourselection.hourobjects.hourobject_helpers import (
    async_update_interim_lists,
)
from ....models.hourselection.day_types import DayTypes

_LOGGER = logging.getLogger(__name__)


class HourSelectionService:
    def __init__(self, parent: Hoursselection, base_mock_hour: int | None = None):
        self.parent = parent
        self._mock_hour = base_mock_hour
        self._mock_day: int | None = None
        self.preserve_interim: bool = False
        self._midnight_touched: bool = False

    async def async_update(self) -> None:
        if all(
            [
                len(self.parent.model.prices_today) == 0,
                len(self.parent.model.prices_tomorrow) == 0,
            ]
        ):
            return
        if self.preserve_interim:
            await self.async_change_midnight_data()
        else:
            self._midnight_touched = False
            today = await self.async_update_per_day(
                prices=self.parent.model.prices_today, day_type=DayTypes.Today
            )
            tomorrow = await self.async_update_per_day(
                prices=self.parent.model.prices_tomorrow, day_type=DayTypes.Tomorrow
            )
            hours, hours_tomorrow = await self.async_interim_day_update(today, tomorrow)

            self.parent.model.hours.hours_today = (
                await self.async_add_remove_limited_hours(
                    hours, day_type=DayTypes.Today
                )
            )
            self.parent.model.hours.hours_tomorrow = (
                await self.async_add_remove_limited_hours(
                    hours_tomorrow, day_type=DayTypes.Tomorrow
                )
            )
            hour = await self.async_set_hour()
            await self.parent.model.hours.async_update_hour_lists(hour=hour)

    async def async_change_midnight_data(self) -> None:
        if self.parent.model.prices_tomorrow == []:
            if not self._midnight_touched:
                self._midnight_touched = (
                    await self.parent.model.hours.async_touch_midnight()
                )
        else:
            self.preserve_interim = False
            await self.async_update()

    async def async_update_per_day(
        self, prices: list, day_type: DayTypes, range_start: int = 0
    ) -> HourObject:
        _LOGGER.debug(f"Updating {day_type.name}")
        pricedict = {}
        if prices is not None and len(prices) > 1:
            pricedict = await async_create_dict(prices)
            normalized_pricedict = await async_create_dict(normalize_prices(prices))
            if stat.stdev(prices) > 0.05:
                ranked = await async_create_cautions(
                    pricedict,
                    normalized_pricedict,
                    self.parent.cautionhour_type_enum,
                    range_start,
                    self.parent.model.adjusted_average,
                    self.parent.model.options.blocknocturnal,
                )
                ready_hours = await self.async_determine_hours(ranked, prices)
                ret = HourObject(
                    nh=ready_hours.nh,
                    ch=ready_hours.ch,
                    dyn_ch=ready_hours.dyn_ch,
                    pricedict=pricedict,
                )
            else:
                """the price curve is too flat to determine hours"""
                ret = HourObject(nh=[], ch=[], dyn_ch={}, pricedict=pricedict)
            ret.offset_dict = await async_get_offset_dict(normalized_pricedict)
            return ret
        return HourObject([], [], {})

    async def async_add_remove_limited_hours(
        self, hours: HourObject, day_type: DayTypes
    ) -> HourObject:
        """Removes cheap hours and adds expensive hours set by user limitation"""
        if hours is None or all(
            [len(hours.nh) == 0, len(hours.ch) == 0, len(hours.dyn_ch) == 0]
        ):
            return HourObject(offset_dict=hours.offset_dict, pricedict=hours.pricedict)
        match day_type:
            case DayTypes.Today | DayTypes.Interim:
                await hours.async_add_expensive_hours(
                    self.parent.model.options.absolute_top_price
                )
            case DayTypes.Tomorrow:
                _top = await self.parent.model.options.async_add_tomorrow_to_top_price(
                    self.parent.prices_tomorrow, self._mock_day
                )
                await hours.async_add_expensive_hours(_top)
        return await hours.async_remove_cheap_hours(self.parent.model.options.min_price)

    async def async_determine_hours(self, price_list: dict, prices: list) -> HourObject:
        ret = HourObject([], [], {})
        try:
            for p in price_list:
                if price_list[p]["force_non"] is True:
                    ret.nh.append(p)
                elif await async_should_be_cautionhour(
                    price_list[p],
                    prices,
                    self.parent.model.current_peak,
                    self.parent.model.options.cautionhour_type,
                ):
                    ret.ch.append(p)
                    ret.dyn_ch[p] = round(
                        await async_set_charge_allowance(
                            price_list[p]["permax"],
                            self.parent.model.options.cautionhour_type,
                        ),
                        2,
                    )
                else:
                    ret.nh.append(p)
        except IndexError as e:
            _LOGGER.error(f"Error in determine hours: {e}")
        return ret

    async def async_interim_day_update(
        self, today: HourObject, tomorrow: HourObject
    ) -> Tuple[HourObject, HourObject]:
        """Updates the non- and caution-hours with an adjusted mean of 14h - 13h today-tomorrow to get a more sane nightly curve."""
        if len(self.parent.model.prices_tomorrow) < 23:
            self.preserve_interim = False
            return today, tomorrow

        hour = await self.async_set_hour()
        # hour = len(self.parent.prices) - 10
        # negative_hour = hour * -1
        negative_hour = (len(self.parent.prices) - hour) * -1
        print(hour)
        print(negative_hour)
        pricelist = self.parent.model.prices_today[hour::]
        pricelist[len(pricelist) :] = self.parent.model.prices_tomorrow[0:hour]
        new_hours = await self.async_update_per_day(
            prices=pricelist, range_start=hour, day_type=DayTypes.Interim
        )

        today = await async_update_interim_lists(
            range(hour, len(self.parent.prices)), today, new_hours, hour
        )
        tomorrow = await async_update_interim_lists(
            range(0, hour), tomorrow, new_hours, negative_hour
        )
        self.preserve_interim = True
        return today, tomorrow

    # helpers
    def set_hour(self, testhour: int | None = None) -> int:
        return (
            testhour
            if testhour is not None
            else self._mock_hour
            if self._mock_hour is not None
            else datetime.now().hour
        )

    async def async_set_hour(self, testhour: int | None = None) -> int:
        return (
            testhour
            if testhour is not None
            else self._mock_hour
            if self._mock_hour is not None
            else datetime.now().hour
        )

    async def async_set_day(self, day: int | None = None):
        self._mock_day = day if day is not None else datetime.now().day
